import random

import pandas as pd


def tortilla_message() -> str:
    """Get a random tortilla message"""

    tortilla_messages = [
        "Making a tortilla",
        "Making a tortilla 🫓",
        "Cooking a tortilla",
        "Making a tortilla 🫓",
        "Working on a tortilla",
        "Working on a tortilla 🫓",
        "Rolling out a tortilla",
        "Rolling out a tortilla 🫓",
        "Baking a tortilla",
        "Baking a tortilla 🫓",
        "Grilling a tortilla",
        "Grilling a tortilla 🫓",
        "Toasting a tortilla",
        "Toasting a tortilla 🫓",
    ]

    # Randomly accessing a message
    random_message = random.choice(tortilla_messages)
    return random_message


def human2bytes(size_str: str) -> int:
    """
    Converts a human-readable size string (e.g., "100MB") into bytes.
    Supported units: KB, MB, GB, TB, PB.
    """
    units = {"KB": 10**3, "MB": 10**6, "GB": 10**9, "TB": 10**12, "PB": 10**15}
    size_str = size_str.strip().upper()

    for unit, multiplier in units.items():
        if size_str.endswith(unit):
            try:
                value = float(size_str[: -len(unit)].strip())
                return int(value * multiplier)
            except ValueError:
                raise ValueError(f"Invalid size value in '{size_str}'.")
    raise ValueError(
        f"Unsupported unit in '{size_str}'. Supported units are: {', '.join(units.keys())}."
    )


def bytes2human(size_bytes: int) -> str:
    """
    Converts a size in bytes into a human-readable size string.
    Automatically selects the largest appropriate unit.
    """
    units = [
        ("PB", 10**15),
        ("TB", 10**12),
        ("GB", 10**9),
        ("MB", 10**6),
        ("KB", 10**3),
    ]

    for unit, multiplier in units:
        if size_bytes >= multiplier:
            value = size_bytes / multiplier
            return f"{value:.2f}{unit}"
    return f"{size_bytes}B"  # Bytes if smaller than 1KB


def group_dataframe_by_size(
    metadata: pd.DataFrame, chunk_size: int
) -> list[pd.DataFrame]:
    """
    Aggregate rows of a DataFrame into groups, where the total size of each group
    does not exceed a given chunk size.

    Args:
        metadata (pd.DataFrame): A DataFrame containing metadata.
        chunk_size (int): Maximum total size for each group.

    Returns:
        list[pd.DataFrame]: A list of DataFrames, each containing a group of rows.
    """
    groups = []
    current_group = []
    current_sum = 0

    for index, row in metadata.iterrows():
        size = row["tortilla:length"]
        if current_sum + size <= chunk_size:
            current_group.append(index)
            current_sum += size
        else:
            groups.append(metadata.loc[current_group])
            current_group = [index]
            current_sum = size

    if current_group:  # Add the last group if it has rows
        groups.append(metadata.loc[current_group])

    return groups



import datetime
import pathlib
from typing import Optional, Union

import pandas as pd
import pydantic

from pytortilla.datamodel import utils


class STAC(pydantic.BaseModel):
    """SpatioTemporal Asset Catalog (STAC) metadata."""

    crs: str
    raster_shape: tuple[int, int]
    geotransform: tuple[float, float, float, float, float, float]
    centroid: Optional[str] = None
    time_start: Union[datetime.datetime, int]
    time_end: Optional[Union[datetime.datetime, int]] = None

    @pydantic.model_validator(mode="after")
    def check_times(cls, values):
        """Validates that the time_start is before time_end."""
        # If time_start is a datetime object, convert it to a timestamp
        if isinstance(values.time_start, datetime.datetime):
            values.time_start = values.time_start.timestamp()

        # If time_end is a datetime object, convert it to a timestamp
        if values.time_end is not None:
            if isinstance(values.time_end, datetime.datetime):
                values.time_end = values.time_end.timestamp()

            if values.time_start > values.time_end:
                raise ValueError(
                    f"Invalid times: {values.time_start} > {values.time_end}"
                )

        return values


class RAI(pydantic.BaseModel):
    """Metadata for Responsible AI (RAI) objectives."""

    populationdensity: Optional[Union[int, float]] = None
    female: Optional[Union[int, float]] = None
    womenreproducibleage: Optional[Union[int, float]] = None
    children: Optional[Union[int, float]] = None
    youth: Optional[Union[int, float]] = None
    elderly: Optional[Union[int, float]] = None


class Sample(pydantic.BaseModel):
    """A sample with STAC and RAI metadata."""

    id: str
    file_format: utils.GDAL_FILES
    path: pathlib.Path
    stac_data: Optional[STAC] = None
    rai_data: Optional[RAI] = None

    class Config:
        extra = "allow"

    @pydantic.field_validator("path")
    def check_path(cls, value):
        """
        Validates that the provided path exists.
        """
        if not value.exists():
            raise ValueError(f"{value} does not exist.")
        return value

    def export_metadata(self):
        """
        Exports metadata as a dictionary, including STAC and RAI attributes, and any extra fields.
        """
        # Gather additional metadata (extra fields not defined explicitly)
        extra_metadata = self.model_dump(
            exclude={"id", "path", "stac_data", "rai_data"}, by_alias=True
        )

        # If crs, raster_shape and geotransform are not provided, then create the stac:centroid
        if self.stac_data is not None:
            if self.stac_data.centroid is None:
                if (
                    self.stac_data.crs is not None
                    and self.stac_data.geotransform is not None
                    and self.stac_data.raster_shape is not None
                ):
                    self.stac_data.centroid = utils.raster_centroid(
                        crs=self.stac_data.crs,
                        geotransform=self.stac_data.geotransform,
                        raster_shape=self.stac_data.raster_shape,
                    )

        # Merge all metadata into a single dictionary
        metadata = {
            "internal:path": self.path.resolve().as_posix(),
            "tortilla:id": self.id,
            "tortilla:file_format": self.file_format,
            "tortilla:offset": 0,
            "tortilla:length": self.path.stat().st_size,
            "stac:crs": self.stac_data.crs if self.stac_data is not None else None,
            "stac:geotransform": (
                self.stac_data.geotransform if self.stac_data is not None else None
            ),
            "stac:raster_shape": (
                self.stac_data.raster_shape if self.stac_data is not None else None
            ),
            "stac:time_start": (
                self.stac_data.time_start if self.stac_data is not None else None
            ),
            "stac:time_end": (
                self.stac_data.time_end if self.stac_data is not None else None
            ),
            "stac:centroid": (
                self.stac_data.centroid if self.stac_data is not None else None
            ),
            "rai:populationdensity": (
                self.rai_data.populationdensity if self.rai_data is not None else None
            ),
            "rai:female": self.rai_data.female if self.rai_data is not None else None,
            "rai:womenreproducibleage": (
                self.rai_data.womenreproducibleage
                if self.rai_data is not None
                else None
            ),
            "rai:children": (
                self.rai_data.children if self.rai_data is not None else None
            ),
            "rai:youth": self.rai_data.youth if self.rai_data is not None else None,
            "rai:elderly": self.rai_data.elderly if self.rai_data is not None else None,
        }

        # Remove None values
        metadata = {k: v for k, v in metadata.items() if v is not None}

        return {**metadata, **extra_metadata}


class Samples(pydantic.BaseModel):
    samples: list[Sample]

    @pydantic.model_validator(mode="after")
    def check_samples(cls, values):
        """
        Validates that the samples have unique IDs.
        """
        ids = [sample.id for sample in values.samples]
        if len(ids) != len(set(ids)):
            raise ValueError("The samples must have unique IDs.")
        return values

    def export_metadata(self):
        """
        Exports metadata for all samples in the collection.
        """
        return pd.DataFrame([sample.export_metadata() for sample in self.samples])



import concurrent.futures
import mmap
import os
import pathlib
from typing import List, Union, Callable

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import tqdm

from pytortilla.create import utils
from pytortilla.datamodel.main import Samples


def create(
    samples: Samples,
    output: Union[str, pathlib.Path],
    nworkers: int = min(4, os.cpu_count()),
    chunk_size: str = "20GB",
    chunk_size_iter: str = "100MB",
    tortilla_message: Callable[[], str] = utils.tortilla_message,
    quiet: bool = False
) -> Union[pathlib.Path, List[pathlib.Path]]:
    """Create a tortilla 🫓

    A tortilla is a new simple format for storing files
    optimized for partial & very fast random access.

    Args:
        samples (Samples): The list of samples to be included in
            the tortilla file. The Sample objects must have a unique
            `id` field.
        output (Union[str, pathlib.Path]): The path where the tortilla
            file will be saved.
        nworkers (int, optional): The number of workers to use when writing
            the tortilla. Defaults to 4. As the writing process is I/O bound,
            the number of workers should be significantly lower than the number
            of cores in the machine.
        chunk_size (str, optional): Avoid large tortilla files by splitting
            the data into chunks. By default, if the number of samples exceeds
            20GB, the data will be split into chunks of 20GB. The partitions
            can be read independently, as each partition contains the metadata
            of the samples it contains.
        chunk_size_iter (int, optional): The writting chunk size. By default,
            it is 100MB. Faster computers can use a larger chunk size.
        tortilla_message (Callable[[], str], optional): A function that 
            returns the message to be displayed when creating the tortilla.
            By default, it is utils.tortilla_message which returns a random
            message.
        quiet (bool, optional): If True, the function does not print any
            message. By default, it is False.

    Returns:
        Union[pathlib.Path, List[pathlib.Path]]: The path of the tortilla file.
            If the tortilla file is split into multiple parts, a list of paths
            is returned.
    """

    # Create the output folder if it does not exist
    output = pathlib.Path(output)
    output.parent.mkdir(parents=True, exist_ok=True)

    # From samples to metadata
    metadata: pd.DataFrame = samples.export_metadata()

    # From human-readable to bytes
    chunk_size_iter_bytes: int = utils.human2bytes(chunk_size_iter)
    chunk_size_bytes: int = utils.human2bytes(chunk_size)

    # Aggregate metadata into groups
    metadata_groups: List[pd.DataFrame] = utils.group_dataframe_by_size(
        metadata=metadata, chunk_size=chunk_size_bytes
    )

    # Save the tortilla
    if len(metadata_groups) == 1:
        return create_a_tortilla(
            metadata=metadata_groups[0],
            output=output,
            ndatapartitions=1,
            nworkers=nworkers,
            chunk_size_iter_bytes=chunk_size_iter_bytes,
            tortilla_message=tortilla_message,
            quiet=quiet,
        )

    # Save the tortilla in parts
    paths = []
    for idx, metadata in enumerate(metadata_groups):
        output_suffix: str = output.suffix
        paths.append(
            create_a_tortilla(
                metadata=metadata,
                output=output.with_suffix(f".{str(idx).zfill(4)}.part.{output_suffix}"),
                ndatapartitions=len(metadata_groups),
                nworkers=nworkers,
                chunk_size_iter_bytes=chunk_size_iter_bytes,
                tortilla_message=tortilla_message,
                quiet=quiet,
            )
        )
    return paths


def create_a_tortilla(
    metadata: pd.DataFrame,
    output: Union[str, pathlib.Path],
    ndatapartitions: int,
    nworkers: int,
    chunk_size_iter_bytes: int,
    tortilla_message: Callable[[], str],
    quiet: bool = False,
) -> pathlib.Path:
    """Create a SINGLE tortilla file 🫓"""

    # Set the Magic Number (MB)
    # Don't forget 3.oct.11
    MB: bytes = b"#y"

    # Estimate the new offset
    metadata.loc[:, "tortilla:offset"] = (
        metadata["tortilla:length"].shift(1, fill_value=0).cumsum() + 200
    )

    # Number of bytes in the DATA pile
    bytes_counter: int = (
        metadata.iloc[-1]["tortilla:length"] + metadata.iloc[-1]["tortilla:offset"]
    )

    # Drop the internal:path field (the paths to the original files)
    internal_path: List[str] = metadata["internal:path"].tolist()
    metadata.drop(columns=["internal:path"], inplace=True)

    # Create an in-memory Parquet file with BufferOutputStream
    with pa.BufferOutputStream() as sink:
        pq.write_table(
            pa.Table.from_pandas(metadata),
            sink,
            compression="zstd",  # Highly efficient codec
            compression_level=22,  # Maximum compression for Zstandard
            use_dictionary=False,  # Optimizes for repeated values
        )
        # return a blob of the in-memory Parquet file as bytes
        # This is the FOOTER metadata
        FOOTER: bytes = sink.getvalue().to_pybytes()

    # Define the FOOTER length and offset
    FL: bytes = len(FOOTER).to_bytes(8, "little")
    FO: bytes = int(bytes_counter).to_bytes(8, "little")

    # Get the total size of the files
    total_size: int = len(FOOTER) + bytes_counter

    # If the folder does not exist, create it
    output = pathlib.Path(output)
    output.parent.mkdir(parents=True, exist_ok=True)

    # Create an empty file
    with open(output, "wb") as f:
        f.truncate(total_size)

    # Define the function to write into the main file
    def write_file(file, offset, length):
        """read the file in chunks"""
        with open(file, "rb") as g:
            while True:
                chunk = g.read(chunk_size_iter_bytes)
                if not chunk:
                    break
                mm[offset : (offset + length)] = chunk

    # Cook the tortilla 🫓
    with open(output, "r+b") as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_WRITE) as mm:
            # Write the magic number (MB)
            mm[:2] = MB

            # Write the FOOTER offset (FO)
            mm[2:10] = FO

            # Write the FOOTER length (FL)
            mm[10:18] = FL

            # Write the number of data partitions (DP)
            mm[18:26] = ndatapartitions.to_bytes(8, "little")

            # Write the free space (FP)
            mm[26:200] = b"\0" * 174

            # Write the DATA pile
            message = tortilla_message()
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=nworkers
            ) as executor:

                # Submit the files to be written
                futures = []
                for path, offset, length in zip(
                    internal_path,
                    metadata["tortilla:offset"],
                    metadata["tortilla:length"],
                ):
                    futures.append(executor.submit(write_file, path, offset, length))

                # Wait for all futures to complete
                if not quiet:
                    list(
                        tqdm.tqdm(
                            concurrent.futures.as_completed(futures),
                            total=len(futures),
                            desc=message,
                            unit="file",
                        )
                    )
                else:
                    concurrent.futures.wait(futures)

            # Write the FOOTER
            mm[bytes_counter : (bytes_counter + len(FOOTER))] = FOOTER

    return output



from typing import Literal, Tuple

from pyproj import CRS, Transformer


def raster_centroid(
    crs: str,
    geotransform: Tuple[float, float, float, float, float, float],
    raster_shape: Tuple[int, int],
) -> str:
    """
    Calculate the centroid of a raster in EPSG:4326 and return a WKT string.

    Args:
        crs (str): The raster's Coordinate Reference System (e.g., "EPSG:32633").
        geotransform (Tuple[float, float, float, float, float, float]): The
            geotransform of the raster following the GDAL convention:
            (
                top left x,
                x resolution,
                x rotation,
                top left y,
                y rotation,
                y resolution
            )
        raster_shape (Tuple[int, int]): The shape of the raster as (rows, columns).

    Returns:
        str: The centroid coordinates in EPSG:4326 as a WKT string.
    """
    # Extract geotransform parameters
    origin_x, pixel_width, _, origin_y, _, pixel_height = geotransform
    rows, cols = raster_shape

    # Compute raster centroid in the raster CRS
    centroid_x = origin_x + (cols / 2) * pixel_width
    centroid_y = origin_y + (rows / 2) * pixel_height

    # Transform centroid to EPSG:4326 using pyproj
    transformer = Transformer.from_crs(
        CRS.from_string(crs), CRS.from_epsg(4326), always_xy=True
    )
    lon, lat = transformer.transform(centroid_x, centroid_y)

    # precision of 6 decimal places
    lon = round(lon, 6)
    lat = round(lat, 6)

    return f"POINT ({lon} {lat})"


GDAL_FILES = Literal[
    "VRT",
    "DERIVED",
    "GTiff",
    "COG",
    "NITF",
    "RPFTOC",
    "ECRGTOC",
    "HFA",
    "SAR_CEOS",
    "CEOS",
    "JAXAPALSAR",
    "GFF",
    "ELAS",
    "ESRIC",
    "AIG",
    "AAIGrid",
    "GRASSASCIIGrid",
    "ISG",
    "SDTS",
    "DTED",
    "PNG",
    "JPEG",
    "MEM",
    "JDEM",
    "GIF",
    "BIGGIF",
    "ESAT",
    "FITS",
    "BSB",
    "XPM",
    "BMP",
    "DIMAP",
    "AirSAR",
    "RS2",
    "SAFE",
    "PCIDSK",
    "PCRaster",
    "ILWIS",
    "SGI",
    "SRTMHGT",
    "Leveller",
    "Terragen",
    "netCDF",
    "HDF4",
    "HDF4Image",
    "ISIS3",
    "ISIS2",
    "PDS",
    "PDS4",
    "VICAR",
    "TIL",
    "ERS",
    "JP2OpenJPEG",
    "L1B",
    "FIT",
    "GRIB",
    "RMF",
    "WCS",
    "WMS",
    "MSGN",
    "RST",
    "GSAG",
    "GSBG",
    "GS7BG",
    "COSAR",
    "TSX",
    "COASP",
    "R",
    "MAP",
    "KMLSUPEROVERLAY",
    "WEBP",
    "PDF",
    "Rasterlite",
    "MBTiles",
    "PLMOSAIC",
    "CALS",
    "WMTS",
    "SENTINEL2",
    "MRF",
    "PNM",
    "DOQ1",
    "DOQ2",
    "PAux",
    "MFF",
    "MFF2",
    "GSC",
    "FAST",
    "BT",
    "LAN",
    "CPG",
    "NDF",
    "EIR",
    "DIPEx",
    "LCP",
    "GTX",
    "LOSLAS",
    "NTv2",
    "CTable2",
    "ACE2",
    "SNODAS",
    "KRO",
    "ROI_PAC",
    "RRASTER",
    "BYN",
    "NOAA_B",
    "NSIDCbin",
    "ARG",
    "RIK",
    "USGSDEM",
    "GXF",
    "BAG",
    "S102",
    "HDF5",
    "HDF5Image",
    "NWT_GRD",
    "NWT_GRC",
    "ADRG",
    "SRP",
    "BLX",
    "PostGISRaster",
    "SAGA",
    "XYZ",
    "HF2",
    "OZI",
    "CTG",
    "ZMap",
    "NGSGEOID",
    "IRIS",
    "PRF",
    "EEDAI",
    "DAAS",
    "SIGDEM",
    "HEIF",
    "TGA",
    "OGCAPI",
    "STACTA",
    "STACIT",
    "GPKG",
    "OpenFileGDB",
    "CAD",
    "PLSCENES",
    "NGW",
    "GenBin",
    "ENVI",
    "EHdr",
    "ISCE",
    "Zarr",
    "HTTP",
    "BYTES",
    "TORTILLA",
]
